@@ -30,6 +30,7 @@ from socket import gethostname
 import traceback
 
 import slack
+import rocket
 
 
 # How slack expresses mentioning users
@@ -37,7 +38,7 @@ _MENTIONS_REGEXP = re.compile(r'<@([0-9A-Za-z]+)>')
 _CHANNEL_MENTIONS_REGEXP = re.compile(r'<#[A-Z0-9]+\|([A-Z0-9\-a-z]+)>')
 
 
-_SUBSTITUTIONS = [
+_SLACK_SUBSTITUTIONS = [
     ('&amp;', '&'),
     ('&gt;', '>'),
     ('&lt;', '<'),
@@ -62,24 +63,34 @@ class Replies(Enum):
     ERR_ERRONEUSNICKNAME = 432
 
 
+class Provider(Enum):
+    SLACK = 0
+    ROCKETCHAT = 1
+
+
 #: Inactivity days to hide a MPIM
 MPIM_HIDE_DELAY = datetime.timedelta(days=50)
 
 
 class Client:
-    def __init__(self, s, sl_client, *, nouserlist=False, autojoin=False):
+    def __init__(self, s, sl_client, nouserlist: bool, autojoin: bool, provider: Provider):
         self.nick = b''
         self.username = b''
         self.realname = b''
         self.parted_channels = set()  # type: Set[bytes]
         self.hostname = gethostname().encode('utf8')
 
+        self.provider = provider
         self.s = s
         self.sl_client = sl_client
-
         self.nouserlist = nouserlist
         self.autojoin = autojoin
 
+        if self.provider == Provider.SLACK:
+            self.substitutions = _SLACK_SUBSTITUTIONS
+        else:
+            self.substitutions: List[Tuple[str,str]] = []
+
     def _nickhandler(self, cmd: bytes) -> None:
         _, nick = cmd.split(b' ', 1)
         self.nick = nick.strip()
@@ -177,18 +188,30 @@ class Client:
         _, dest, msg = cmd.split(b' ', 2)
         if msg.startswith(b':'):
             msg = msg[1:]
+
+        # Handle sending "/me does something"
+        # b'PRIVMSG #much_private :\x01ACTION saluta tutti\x01'
+        if msg.startswith(b'\x01ACTION ') and msg.endswith(b'\x01'):
+            action = True
+            _, msg = msg.split(b' ', 1)
+            msg = msg[:-1]
+        else:
+            action = False
+
         message = self._addmagic(msg.decode('utf8'))
 
         if dest.startswith(b'#'):
             self.sl_client.send_message(
                 self.sl_client.get_channel_by_name(dest[1:].decode()).id,
-                message
+                message,
+                action,
             )
         else:
             try:
                 self.sl_client.send_message_to_user(
                     self.sl_client.get_user_by_name(dest.decode()).id,
-                    message
+                    message,
+                    action,
                 )
             except:
                 print('Impossible to find user ', dest)
@@ -238,6 +261,11 @@ class Client:
         response = Replies.RPL_NOWAWAY if is_away else Replies.RPL_UNAWAY
         self._sendreply(response, 'Away status changed')
 
+    def _topichandler(self, cmd: bytes) -> None:
+        _, channel, topic = cmd.split(b' ', 2)
+        channel = self.sl_client.get_channel_by_name(channel.decode()[1:])
+        self.sl_client.topic(channel, topic.decode()[1:])
+
     def _whohandler(self, cmd: bytes) -> None:
         _, name = cmd.split(b' ', 1)
         if not name.startswith(b'#'):
@@ -249,8 +277,11 @@ class Client:
             return
 
         for i in self.sl_client.get_members(channel.id):
-            user = self.sl_client.get_user(i)
-            self._sendreply(Replies.RPL_WHOREPLY, '0 %s' % user.real_name, [name, user.name, '127.0.0.1', self.hostname, user.name, 'H'])
+            try:
+                user = self.sl_client.get_user(i)
+                self._sendreply(Replies.RPL_WHOREPLY, '0 %s' % user.real_name, [name, user.name, '127.0.0.1', self.hostname, user.name, 'H'])
+            except:
+                pass
         self._sendreply(Replies.RPL_ENDOFWHO, 'End of WHO list', [name])
 
     def sendmsg(self, from_: bytes, to: bytes, message: bytes) -> None:
@@ -265,20 +296,29 @@ class Client:
         Adds magic codes and various things to
         outgoing messages
         """
-        for i in _SUBSTITUTIONS:
+        for i in self.substitutions:
             msg = msg.replace(i[1], i[0])
-        msg = msg.replace('@here', '<!here>')
-        msg = msg.replace('@channel', '<!channel>')
-        msg = msg.replace('@yell', '<!channel>')
-        msg = msg.replace('@shout', '<!channel>')
-        msg = msg.replace('@attention', '<!channel>')
+        if self.provider == Provider.SLACK:
+            msg = msg.replace('@here', '<!here>')
+            msg = msg.replace('@channel', '<!channel>')
+            msg = msg.replace('@yell', '<!channel>')
+            msg = msg.replace('@shout', '<!channel>')
+            msg = msg.replace('@attention', '<!channel>')
+        elif self.provider == Provider.ROCKETCHAT:
+            msg = msg.replace('@yell', '@channel')
+            msg = msg.replace('@shout', '@channel')
+            msg = msg.replace('@attention', '@channel')
 
         # Extremely inefficient code to generate mentions
         # Just doing them client-side on the receiving end is too mainstream
         for username in self.sl_client.get_usernames():
+            if len(username) < 3: continue
             m = re.search(r'\b%s\b' % username, msg)
             if m:
-                msg = msg[0:m.start()] + '<@%s>' % self.sl_client.get_user_by_name(username).id + msg[m.end():]
+                if self.provider == Provider.SLACK:
+                    msg = msg[0:m.start()] + '<@%s>' % self.sl_client.get_user_by_name(username).id + msg[m.end():]
+                elif self.provider == Provider.ROCKETCHAT:
+                    msg = msg[0:m.start()] + f'@{username}' + msg[m.end():]
         return msg
 
     def parse_message(self, msg: str) -> Iterator[bytes]:
@@ -298,36 +338,39 @@ class Client:
                 )
 
             # Replace all channel mentions
-            while True:
-                mention = _CHANNEL_MENTIONS_REGEXP.search(i)
-                if not mention:
-                    break
-                i = (
-                    i[0:mention.span()[0]] +
-                    '#' +
-                    mention.groups()[0] +
-                    i[mention.span()[1]:]
-                )
+            if self.provider == Provider.SLACK:
+                while True:
+                    mention = _CHANNEL_MENTIONS_REGEXP.search(i)
+                    if not mention:
+                        break
+                    i = (
+                        i[0:mention.span()[0]] +
+                        '#' +
+                        mention.groups()[0] +
+                        i[mention.span()[1]:]
+                    )
 
-            for s in _SUBSTITUTIONS:
+            for s in self.substitutions:
                 i = i.replace(s[0], s[1])
 
             encoded = i.encode('utf8')
 
-            encoded = encoded.replace(b'<!here>', b'yelling [%s]' % self.nick)
-            encoded = encoded.replace(b'<!channel>', b'YELLING LOUDER [%s]' % self.nick)
+            if self.provider == Provider.SLACK:
+                encoded = encoded.replace(b'<!here>', b'yelling [%s]' % self.nick)
+                encoded = encoded.replace(b'<!channel>', b'YELLING LOUDER [%s]' % self.nick)
+            elif self.provider == Provider.ROCKETCHAT:
+                encoded = encoded.replace(b'@here', b'yelling [%s]' % self.nick)
+                encoded = encoded.replace(b'@channel', b'YELLING LOUDER [%s]' % self.nick)
 
             yield encoded
 
 
-    def _message(self, sl_ev: Union[slack.Message, slack.MessageDelete, slack.MessageBot], prefix: str=''):
+    def _message(self, sl_ev: Union[slack.Message, slack.MessageDelete, slack.MessageBot, slack.ActionMessage], prefix: str=''):
         """
         Sends a message to the irc client
         """
         if hasattr(sl_ev, 'user'):
             source = self.sl_client.get_user(sl_ev.user).name.encode('utf8')  # type: ignore
-            if source == self.nick:
-                return
         else:
             source = b'bot'
         try:
@@ -341,6 +384,8 @@ class Client:
             # Ignoring messages, channel was left on IRC
             return
         for msg in self.parse_message(prefix + sl_ev.text):
+            if isinstance(sl_ev, slack.ActionMessage):
+                msg = b'\x01ACTION ' + msg + b'\x01'
             self.sendmsg(
                 source,
                 dest,
@@ -368,6 +413,8 @@ class Client:
             self._message(sl_ev, '[deleted]')
         elif isinstance(sl_ev, slack.Message):
             self._message(sl_ev)
+        elif isinstance(sl_ev, slack.ActionMessage):
+            self._message(sl_ev)
         elif isinstance(sl_ev, slack.MessageEdit):
             if sl_ev.is_changed:
                 self._message(sl_ev.diffmsg)
@@ -378,6 +425,8 @@ class Client:
             self._message(f.announce())
         elif isinstance(sl_ev, slack.Join):
             self._joined(sl_ev)
+        elif isinstance(sl_ev, slack.TopicChange):
+            self._sendreply(Replies.RPL_TOPIC, sl_ev.topic, ['#' + self.sl_client.get_channel(sl_ev.channel).name])
 
     def command(self, cmd: bytes) -> None:
         if b' ' in cmd:
@@ -396,11 +445,11 @@ class Client:
             b'MODE': self._modehandler,
             b'PART': self._parthandler,
             b'AWAY': self._awayhandler,
+            b'TOPIC': self._topichandler,
             b'sendfile': self._sendfilehandler,
             #QUIT
             #CAP LS
             #USERHOST
-            #Unknown command:  b'TOPIC #cama :titolo del canale'
             #Unknown command:  b'whois TAMARRO'
         }
 
@@ -432,6 +481,8 @@ def main() -> None:
     parser.add_argument('-o', '--override', action='store_true',
                                 dest='overridelocalip', required=False,
                                 help='allow non 127. addresses, this is potentially dangerous')
+    parser.add_argument('--rc-url', type=str, action='store', dest='rc_url', default=None, required=False,
+                                help='The rocketchat URL. Setting this changes the mode from slack to rocketchat')
 
     args = parser.parse_args()
     # Exit if their chosden ip isn't local. User can override with -o if they so dare
@@ -454,7 +505,12 @@ def main() -> None:
         except (FileNotFoundError, PermissionError):
             exit(f'Unable to open the token file {args.tokenfile}')
 
-    sl_client = slack.Slack(token)
+    if args.rc_url:
+        sl_client = rocket.Rocket(args.rc_url, token)  # type: Union[slack.Slack, rocket.Rocket]
+        provider = Provider.ROCKETCHAT
+    else:
+        sl_client = slack.Slack(token)
+        provider = Provider.SLACK
     sl_events = sl_client.events_iter()
     serversocket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
     serversocket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
@@ -465,14 +521,14 @@ def main() -> None:
 
     while True:
         s, _ = serversocket.accept()
-        ircclient = Client(s, sl_client, nouserlist=args.nouserlist, autojoin=args.autojoin)
+        ircclient = Client(s, sl_client, args.nouserlist, args.autojoin, provider)
 
         poller.register(s.fileno(), select.POLLIN)
         if sl_client.fileno is not None:
             poller.register(sl_client.fileno, select.POLLIN)
 
         # Main loop
-        timeout = 7
+        timeout = 2
         while True:
             s_event = poller.poll(timeout)  # type: List[Tuple[int,int]]
             sl_event = next(sl_events)
