@@ -18,7 +18,7 @@
 
 import datetime
 from functools import lru_cache
-from time import sleep
+from time import sleep, time
 from typing import *
 
 from slackclient import SlackClient
@@ -57,6 +57,7 @@ class Response(NamedTuple):
     """
     ok: bool
     headers: Dict[str, str]
+    ts: Optional[float] = None
 
 
 class Topic(NamedTuple):
@@ -114,6 +115,10 @@ class Message(NamedTuple):
     text: str
 
 
+class ActionMessage(Message):
+    pass
+
+
 class MessageEdit(NamedTuple):
     previous: Message
     current: Message
@@ -204,10 +209,17 @@ class Join(NamedTuple):
     channel: str
 
 
+class TopicChange(NamedTuple):
+    topic: str
+    channel: str
+    user: str
+
 SlackEvent = Union[
+    TopicChange,
     MessageDelete,
     MessageEdit,
     Message,
+    ActionMessage,
     MessageBot,
     FileShared,
     Join,
@@ -223,6 +235,7 @@ class Slack:
         self._get_members_cache = {}  # type: Dict[str, Set[str]]
         self._get_members_cache_cursor = {}  # type: Dict[str, Optional[str]]
         self._internalevents = []  # type: List[SlackEvent]
+        self._sent_by_self = set()  # type: Set[float]
 
     def away(self, is_away: bool) -> None:
         """
@@ -234,6 +247,12 @@ class Slack:
         if not response.ok:
             raise ResponseException(response)
 
+    def topic(self, channel: Channel, topic: str) -> None:
+        r = self.client.api_call('conversations.setTopic', channel=channel.id, topic=topic)
+        response = load(r, Response)
+        if not response.ok:
+            raise ResponseException(response)
+
     def get_members(self, id_: str) -> Set[str]:
         """
         Returns the list (as a set) of users in a channel.
@@ -311,6 +330,19 @@ class Slack:
     def fileno(self) -> Optional[int]:
         return self.client.fileno
 
+    def get_im(self, im_id: str) -> Optional[IM]:
+        if not im_id.startswith('D'):
+            return None
+        for uid, imid in self._imcache.items():
+            if im_id == imid:
+                return IM(user=uid, id=imid)
+
+        for im in self.get_ims():
+            self._imcache[im.user] = im.id
+            if im.id == im_id:
+                return im;
+        return None
+
     def get_ims(self) -> List[IM]:
         """
         Returns a list of the IMs
@@ -394,22 +426,39 @@ class Slack:
             return
         raise ResponseException(response)
 
-    def send_message(self, channel_id: str, msg: str) -> None:
+    def _triage_sent_by_self(self) -> None:
+        """
+        Clear all the old leftovers in
+        _sent_by_self
+        """
+        r = []
+        for i in self._sent_by_self:
+            if time() - i >= 10:
+                r.append(i)
+        for i in r:
+            self._sent_by_self.remove(i)
+
+    def send_message(self, channel_id: str, msg: str, action: bool) -> None:
         """
         Send a message to a channel or group or whatever
         """
+        if action:
+            api = 'chat.meMessage'
+        else:
+            api = 'chat.postMessage'
         r = self.client.api_call(
-            "chat.postMessage",
+            api,
             channel=channel_id,
             text=msg,
             as_user=True,
         )
         response = load(r, Response)
-        if response.ok:
+        if response.ok and response.ts:
+            self._sent_by_self.add(response.ts)
             return
         raise ResponseException(response)
 
-    def send_message_to_user(self, user_id: str, msg: str):
+    def send_message_to_user(self, user_id: str, msg: str, action: bool):
         """
         Send a message to a user, pass the user id
         """
@@ -444,7 +493,7 @@ class Slack:
 
             self._imcache[user_id] = channel_id
 
-        self.send_message(channel_id, msg)
+        self.send_message(channel_id, msg, action)
 
     def events_iter(self) -> Iterator[Optional[SlackEvent]]:
         """
@@ -476,9 +525,25 @@ class Slack:
                 t = event.get('type')
                 subt = event.get('subtype')
 
+                ts = float(event.get('ts', 0))
+                if ts in self._sent_by_self:
+                    self._sent_by_self.remove(ts)
+                    continue
+
                 try:
-                    if t == 'message' and not subt:
-                        yield _loadwrapper(event, Message)
+                    if t == 'message' and (not subt or subt == 'me_message'):
+                        msg = _loadwrapper(event, Message)
+
+                        # In private chats, pretend that my own messages
+                        # sent from another client actually come from
+                        # the other user, and prepend them with "I say: "
+                        im = self.get_im(msg.channel)
+                        if im and im.user != msg.user:
+                            msg = Message(user=im.user, text='I say: ' + msg.text, channel=im.id)
+                        if subt == 'me_message':
+                            yield ActionMessage(*msg)
+                        else:
+                            yield msg
                     elif t == 'message' and subt == 'slackbot_response':
                         yield _loadwrapper(event, Message)
                     elif t == 'message' and subt == 'message_changed':
@@ -488,6 +553,8 @@ class Slack:
                             previous=load(event['previous_message'], Message),
                             current=load(event['message'], Message)
                         )
+                    elif t == 'message' and subt == 'group_topic':
+                        yield _loadwrapper(event, TopicChange)
                     elif t == 'message' and subt == 'message_deleted':
                         event['previous_message']['channel'] = event['channel']
                         ev = _loadwrapper(event['previous_message'], MessageDelete)
@@ -517,4 +584,5 @@ class Slack:
                         print(event)
                 except Exception as e:
                     print('Exception: %s' % e)
+            self._triage_sent_by_self()
             yield None
